import pygame as pg
import math as m
from pygame.colordict import THECOLORS
from typing import Optional, Tuple, Sequence
import pg3d.MatrixMath.matrix as mm
from pg3d.camera import Camera
from pg3d.triangle import Triangle


class App:
    def __init__(
        self,
        dimensions=(1000, 700),
        cam_pos=[0, 0, 0],
        BG_COLOR=(0, 0, 0),
        LINE_COLOR=(255, 255, 255),
        VERTEX_SIZE=2,
        stats=False,
        fullscreen=False,
        mouse_look=False,
    ):
        """
        Initialises the library, creates the projection matrix and creates a camera

        Args:
            dimensions ([tuple], optional): [window dimensions]. Defaults to (1000, 700).
            cam_pos ([list], optional): [position of camrea]. Defaults to [0, 0, 0].
            BG_COLOR ([tuple], optional): [background color]. Defaults to (0, 0, 0).
            LINE_COLOR ([tuple], optional): [color for drawing lines and points]. Defaults to (255, 255, 255).
            VERTEX_SIZE ([int], optional): [size of points]. Defaults to 2.
            stats ([bool], optional): [shows some stats on screen]. Defaults to False.
            fullscreen ([bool], optional): [makes screen fullscreen]. Defaults to False.
            mouse_look ([bool], optional): [use mouse movement too look with camera]. Defaults to False.
        """
        pg.init()

        if fullscreen:
            self.screen = pg.display.set_mode((0, 0), pg.FULLSCREEN, vsync=1)
            self.dimensions = (
                self.width,
                self.height,
            ) = pg.display.get_surface().get_size()
        else:
            self.dimensions = self.width, self.height = dimensions
            self.screen = pg.display.set_mode(self.dimensions, pg.RESIZABLE, vsync=1)

        self.half_width, self.half_height = self.width / 2, self.height / 2
        self.FPS: int = 60
        self.screen = pg.display.set_mode(self.dimensions, pg.RESIZABLE, vsync=1)
        self.clock = pg.time.Clock()
        self.stats = stats

        if mouse_look:
            pg.mouse.set_visible(0)
        self.mouse_look = mouse_look

        self.BG_COLOR = BG_COLOR
        self.LINE_COLOR = LINE_COLOR
        self.VERTEX_SIZE = VERTEX_SIZE

        self.camera = Camera(self, cam_pos)

        self.mesh = []

        self.fov = 90
        self.zf = 1000
        self.zn = 0.1

        m00 = (self.height / self.width) * (1 / m.tan(m.radians(self.fov / 2)))
        m11 = 1 / m.tan(m.radians(self.fov / 2))
        m22 = self.zf / (self.zf - self.zn)
        m32 = -self.zn * (self.zf / (self.zf - self.zn))

        self.projection_matrix = mm.Matrix(
            [[m00, 0, 0, 0], [0, m11, 0, 0], [0, 0, m22, 1], [0, 0, m32, 0]]
        )

    def _update_projection_matrix(self):
        """
        Updates the projection matrix when the values of fov and aspect ratio are changed by the user
        """
        m00 = (self.height / self.width) * (1 / m.tan(m.radians(self.fov / 2)))
        m11 = 1 / m.tan(m.radians(self.fov / 2))
        m22 = self.zf / (self.zf - self.zn)
        m32 = -self.zn * (self.zf / (self.zf - self.zn))

        self.projection_matrix = mm.Matrix(
            [[m00, 0, 0, 0], [0, m11, 0, 0], [0, 0, m22, 1], [0, 0, m32, 0]]
        )

    def _add_point(self, point):
        """
        When a user creates a point object this function is called and adds the point to mesh

        Args:
            point ([Point]): [a point object]
        """
        self.mesh.append([point])

    def _add_triangle(self, triangle):
        self.mesh.append(triangle)

    def _draw(self):
        self.screen.fill(self.BG_COLOR)

        for shape in self.mesh:
            if type(shape) == Triangle:
                shape._project()

            else:
                for point in shape:
                    projected = point._project(
                        self.projection_matrix, self.camera._cam_mat()
                    )

                    if projected is not None:
                        x, y, z = projected

                        if point.vertex == True:
                            pg.draw.circle(
                                self.screen, self.LINE_COLOR, (x, y), self.VERTEX_SIZE
                            )

    def _display_stats(self):
        """
        If self.stats is true, this method will display stats on screen every frame
        """
        font = pg.font.Font("freesansbold.ttf", 10)
        fov = font.render(f"fov = {self.fov}", True, (0, 255, 0))
        fps = font.render(f"fps = {round(self.clock.get_fps())}", True, (0, 255, 0))
        dimensions = font.render(f"dimensions = {self.dimensions}", True, (0, 255, 0))
        self.screen.blit(fov, (5, 5))
        self.screen.blit(fps, (5, 15))
        self.screen.blit(dimensions, (5, 25))

    def _check_events(self):
        for event in pg.event.get():
            if event.type == pg.QUIT:
                exit()

            elif event.type == pg.KEYDOWN and event.key == pg.K_ESCAPE:
                exit()

            elif event.type == pg.MOUSEWHEEL:
                if event.y == 1:
                    self.fov -= 1
                else:
                    self.fov += 1
                self._update_projection_matrix()

            elif event.type == pg.VIDEORESIZE:
                self.screen = pg.display.set_mode((event.w, event.h), pg.RESIZABLE)
                self.dimensions = self.width, self.height = (event.w, event.h)
                self.half_width, self.half_height = self.width / 2, self.height / 2
                self._update_projection_matrix()

            elif (event.type == pg.MOUSEMOTION) and (self.mouse_look == True):
                self.camera._mouse_look(event.rel)

    def run(self):
        """
        Main loop of the library which checks for camera control and other events, and draws and projects the points
        """
        while True:
            self._draw()
            self.camera._movement()
            self._check_events()
            if self.stats:
                self._display_stats()

            if self.mouse_look == True:
                pg.mouse.set_pos((self.half_width, self.half_height))

            pg.display.set_caption(f"{round(self.clock.get_fps())} FPS")
            pg.display.update()
            self.clock.tick(self.FPS)



from pg3d.point import Point
import pygame as pg


class Triangle:
    def __init__(self, app, vertices):
        """
        Creates triangle

        Args:
            app ([App]): [Specify App object]
            vertices ([list[list]]): [list with 3 cartesian coordinates]
        """
        self.points = [
            Point(app, vertices[0], False),
            Point(app, vertices[1], False),
            Point(app, vertices[2], False),
        ]
        self.projected_points = []
        self.app = app
        self.app._add_triangle(self)

    def _project(self):
        """
        Projects triangle
        """
        self.projected_points = []
        for point in self.points:
            projected = point._project(
                self.app.projection_matrix, self.app.camera._cam_mat()
            )
            if projected != None:
                self.projected_points.append(projected)

        self._draw_triangle()

    def _draw_triangle(self):
        """
        Draws triangle
        """
        if len(self.projected_points) == 3:
            a, b, c = self.projected_points
            pg.draw.polygon(
                self.app.screen, self.app.LINE_COLOR, (a[:-1], b[:-1], c[:-1]), 1
            )

    def __getitem__(self, index):
        """AI is creating summary for __getitem__

        Args:
            index ([int]): [index of triangle]

        Returns:
            [Point]: [returns point]
        """
        return self.points[index]



import pg3d.MatrixMath.matrix as mm
import math as m


def rotate_x(angle):
    """
    Rotation matrix on x-axis

    Args:
        angle ([float]): [Angle for rotation]

    Returns:
        [Matrix]: [Creates rotation matrix along x-axis]
    """
    return mm.Matrix(
        [
            [1, 0, 0, 0],
            [0, m.cos(angle), m.sin(angle), 0],
            [0, -m.sin(angle), m.cos(angle), 0],
            [0, 0, 0, 1],
        ]
    )


def rotate_y(angle):
    """
    Rotation matrix on y-axis

    Args:
        angle ([float]): [Angle for rotation]

    Returns:
        [Matrix]: [Creates rotation matrix along y-axis]
    """
    return mm.Matrix(
        [
            [m.cos(angle), 0, -m.sin(angle), 0],
            [0, 1, 0, 0],
            [m.sin(angle), 0, m.cos(angle), 0],
            [0, 0, 0, 1],
        ]
    )


def rotate_z(angle):
    """
    Rotation matrix on z-axis

    Args:
        angle ([float]): [Angle for rotation]

    Returns:
        [Matrix]: [Creates rotation matrix along z-axis]
    """
    return mm.Matrix(
        [
            [m.cos(angle), m.sin(angle), 0, 0],
            [-m.sin(angle), m.cos(angle), 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1],
        ]
    )



import pg3d.MatrixMath.matrix as mm
from pg3d.matrices import rotate_x, rotate_y, rotate_z
import pygame as pg
import math as m


class Camera:
    def __init__(self, app, position, forward=[0, 0, 1], up=[0, 1, 0], right=[1, 0, 0]):
        """
        Args:
            app ([App]): [App instance]
            position ([list]): [Starting position of camera]
            forward ([list], optional): [Orientation along z-axis]. Defaults to [0, 0, 1].
            up ([list], optional): [Orientation along y-axis]. Defaults to [0, 1, 0].
            right ([list], optional): [Orientation along x-axis]. Defaults to [1, 0, 0].
        """
        self.app = app

        self.pos = mm.Matrix([[*position, 1]])
        self.forward = mm.Matrix([[*forward, 1]])
        self.up = mm.Matrix([[*up, 1]])
        self.right = mm.Matrix([[*right, 1]])

        self.speed = 1
        self.angle = m.radians(1)

    def yaw(self, angle):
        """
        Rotates camera directions along y-axis by angle

        Args:
            angle ([float]): [Angle used to rotate camera orientation]
        """
        self.up *= rotate_y(angle)
        self.forward *= rotate_y(angle)
        self.right *= rotate_y(angle)

    def pitch(self, angle):
        """
        Rotates camera directions along x-axis by angle

        Args:
            angle ([float]): [Angle used to rotate camera orientation]
        """
        self.up *= rotate_x(angle)
        self.forward *= rotate_x(angle)
        self.right *= rotate_x(angle)

    def _rot_mat(self):
        """
        Creates rotation matrix

        Returns:
            [Matrix]: [Rotation matrix used to rotate all point in world space around camera]
        """
        fx, fy, fz, fw = self.forward[0]
        rx, ry, rz, rw = self.right[0]
        ux, uy, uz, uw = self.up[0]

        return mm.Matrix(
            [[rx, ux, fx, 0], [ry, uy, fy, 0], [rz, uz, fz, 0], [0, 0, 0, 1]]
        )

    def _trans_mat(self):
        """
        Creates translate matrix

        Returns:
            [Matrix]: [Translation matrix used to put camera at center of 3D space]
        """
        x, y, z, w = self.pos[0]

        return mm.Matrix([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [-x, -y, -z, 1]])

    def _cam_mat(self):
        """
        Creates camera matrix

        Returns:
            [Matrix]: [Camera matrix which is multiplied to all points]
        """
        return self._trans_mat() * self._rot_mat()

    def _movement(self):
        """
        Checks for user input and then performs the necessary rotations and translations
        """
        key = pg.key.get_pressed()

        if key[pg.K_a]:
            self.right = self.speed * self.right
            self.pos = self.pos - self.right
        if key[pg.K_d]:
            self.right = self.speed * self.right
            self.pos = self.pos + self.right
        if key[pg.K_w]:
            self.forward = self.speed * self.forward
            self.pos = self.pos + self.forward
        if key[pg.K_s]:
            self.forward = self.speed * self.forward
            self.pos = self.pos - self.forward
        if key[pg.K_q]:
            self.up = self.speed * self.up
            self.pos = self.pos + self.up
        if key[pg.K_e]:
            self.up = self.speed * self.up
            self.pos = self.pos - self.up

        if key[pg.K_LEFT]:
            self.yaw(-self.angle)
        if key[pg.K_RIGHT]:
            self.yaw(self.angle)
        if key[pg.K_UP]:
            self.pitch(self.angle)
        if key[pg.K_DOWN]:
            self.pitch(-self.angle)

    def _mouse_look(self, rel):
        """
        Moves camera when mouse is moved

        Args:
            rel ([float]): [relative position of mouse]
        """
        x, y = rel
        self.yaw(x / 1000)
        self.pitch(-y / 1000)



from pg3d.triangle import Triangle


class Shape:
    def __init__(self, app, shape, size=1, center=[0, 0, 0]):
        """
        Initialises shape

        Args:
            app ([App]): [Specify app object]
            shape ([str]): [Specify shape]
            size (int, optional): [Size of shape]. Defaults to 1.
            center (list, optional): [Center coordinate of shape]. Defaults to [0, 0, 0].
        """
        self.app = app

        self.center = center
        self.size = size

        triangles, vertices = None, None

        if shape.lower() == "cube":
            vertices, triangles = self._cube()
        elif shape.lower() == "pyramid":
            vertices, triangles = self._pyramid()
        elif shape.lower() == "tetrahedron":
            vertices, triangles = self._tetrahedron()

        if (triangles != None) and (vertices != None):
            for triangle in triangles:
                Triangle(
                    self.app,
                    [
                        vertices[triangle[0]],
                        vertices[triangle[1]],
                        vertices[triangle[2]],
                    ],
                )

    def _cube(self):
        """
        Creates points and vertices of cube

        Returns:
            [tuple[list, list]]: [tuple containing list of vertices and list of triangles]
        """
        x, y, z = self.center
        half_size = self.size / 2

        vertices = [
            [x - half_size, y - half_size, z + half_size],  # Front-bottom-left
            [x + half_size, y - half_size, z + half_size],  # Front-bottom-right
            [x + half_size, y + half_size, z + half_size],  # Front-top-right
            [x - half_size, y + half_size, z + half_size],  # Front-top-left
            [x - half_size, y - half_size, z - half_size],  # Back-bottom-left
            [x + half_size, y - half_size, z - half_size],  # Back-bottom-right
            [x + half_size, y + half_size, z - half_size],  # Back-top-right
            [x - half_size, y + half_size, z - half_size],  # Back-top-left
        ]

        triangles = [
            [0, 1, 2],
            [0, 2, 3],  # Front face
            [1, 5, 6],
            [1, 6, 2],  # Right face
            [3, 2, 6],
            [3, 6, 7],  # Top face
            [4, 5, 1],
            [4, 1, 0],  # Bottom face
            [4, 0, 3],
            [4, 3, 7],  # Left face
            [7, 6, 5],
            [7, 5, 4],  # Back face
        ]

        return vertices, triangles

    def _pyramid(self):
        """
        Creates points and vertices of pyramid

        Returns:
            [tuple[list, list]]: [tuple containing list of vertices and list of triangles]
        """
        x, y, z = self.center
        half_size = self.size / 2
        height = self.size

        vertices = [
            [x - half_size, y + half_size, z - half_size],  # Bottom-back-left
            [x - half_size, y + half_size, z + half_size],  # Bottom-front-left
            [x + half_size, y + half_size, z - half_size],  # Bottom-back-right
            [x + half_size, y + half_size, z + half_size],  # Bottom-front-right
            [x, y - half_size, z],  # Top
        ]

        triangles = [
            [0, 1, 2],
            [0, 2, 3],
            [1, 2, 3],  # Base face
            [0, 4, 3],  # Front-left face
            [1, 4, 0],  # Front-right face
            [2, 4, 1],  # Back-right face
            [3, 4, 2],  # Back-left face
        ]

        return vertices, triangles

    def _tetrahedron(self):
        """
        Creates points and vertices of tetrahedron

        Returns:
            [tuple[list, list]]: [tuple containing list of vertices and list of triangles]
        """
        half_size = self.size / 2
        x, y, z = self.center
        height = (
            self.size * 0.86
        )  # Multiply by 0.86 to adjust height to make it equilateral

        vertices = [
            [x, y - height / 3, z],  # Top
            [x - half_size, y + height / 3, z - half_size],  # Bottom-front-left
            [x + half_size, y + height / 3, z - half_size],  # Bottom-front-right
            [x, y + height / 3, z + half_size],  # Bottom-back
        ]

        triangles = [
            [0, 1, 2],  # Front face
            [0, 2, 3],  # Right face
            [0, 3, 1],  # Left face
            [1, 3, 2],  # Bottom face
        ]

        return vertices, triangles
